import os
import os.path
from typing import List, Union
import logging

from swiftclient.multithreading import OutputManager
from swiftclient.service import SwiftError, SwiftService, SwiftUploadObject


logger = logging.getLogger(__name__)

class Base:
    def __init__(self, username=None, password=None, tenant_name=None,
                 tenant_id=None, region_name=None, user_domain_id=None,
                 user_domain_name=None, auth_url=None, auth_version=None,
                 container=None):
        self.username = username
        self.password = password
        self.tenant_name = tenant_name
        self.tenant_id = tenant_id
        self.region_name = region_name
        self.user_domain_id = user_domain_id
        self.user_domain_name = user_domain_name
        self.auth_url = auth_url
        self.auth_version = auth_version  # TODO: assume 3
        self.container = container

    def get_service(self):
        return SwiftService(options={
            "os_username": self.username,
            "os_password": self.password,
            "os_tenant_name": self.tenant_name,
            "os_tenant_id": self.tenant_id,
            "os_region_name": self.region_name,
            "os_auth_url": self.auth_url,
            "auth_version": self.auth_version,
            "os_user_domain_id": self.user_domain_id,
            "os_user_domain_name": self.user_domain_name,
        })


class Downloader(Base):
    """ Downloader for OpenStack swift object storages
    """
    def download(self, remote_path: os.PathLike, local_path: os.PathLike) -> os.PathLike:
        target_filename = os.path.join(local_path, os.path.basename(remote_path))
        with self.get_service() as swift:
            results = swift.download(
                self.container,
                [remote_path],
                options={
                    'out_file': target_filename
                }
            )

            for result in results:
                if result["success"]:
                    return target_filename
                else:
                    raise Exception('Failed to download %s' % remote_path)


class Uploader(Base):
    """ Uploader for OpenStack swift object storages
    """
    def upload(self, local_path: Union[os.PathLike, List[os.PathLike]], remote_dir: os.PathLike) -> List[os.PathLike]:
        paths = local_path if isinstance(local_path, List) else [local_path]
        remote_paths = [
            os.path.join(
                remote_dir,
                os.path.basename(path)
            )
            for path in paths
        ]
        objects = [
            SwiftUploadObject(
                path,
                object_name=remote_path
            )
            for path, remote_path in zip(paths, remote_paths)
        ]

        max_size = max([os.stat(path).st_size for path in paths])
        options = {}

        # use segment options if file is bigger than 5GB
        if (max_size > 1024 * 1024 * 1024 * 5):
            options['segment_size'] = 2 * 1024 * 1024 * 1024
            options['use_slo'] = True

        with self.get_service() as swift:
            # use container first part of path of container as upload container
            container = self.container or paths[0].partition('/')[0]
            results = swift.upload(container=container, objects=objects, options=options)

            for result in results:
                if result["success"]:
                    if "object" in result:
                        logger.info(
                            "'%s' successfully uploaded." % result["object"]
                        )
                    elif "for_object" in result:
                        logger.debug(
                            "Successfully uploaded '%s' segment '%s'."
                            % (result["for_object"], result["segment_index"])
                        )
                else:
                    logger.error(
                        "'%s' upload failed" % result["error"]
                    )
                    raise Exception('Failed to upload %s' % result["error"])

        return remote_paths
